import asyncio
import json
from typing import Any, Awaitable, Callable, List, Tuple, Union, cast

from eth_account import Account
from eth_account.datastructures import SignedMessage, SignedTransaction
from eth_account.messages import encode_defunct
from eth_account.signers.local import LocalAccount
from eth_typing import Address
from eth_typing.encoding import HexStr
from hexbytes import HexBytes
from web3 import Web3
from web3.contract.contract import Contract, ContractFunction
from web3.gas_strategies import rpc
from web3.types import (
    BlockData,
    BlockIdentifier,
    CallOverride,
    Middleware,
    Nonce,
    TxData,
    TxParams,
    TxReceipt,
    Wei,
)
from websockets.client import connect

from web3client.exceptions import TransactionTooExpensive, Web3ClientException
from web3client.helpers.websockets import parse_notification, subscribe_to_notification


class BaseClient:
    """
    Client to interact with a blockchain, with smart contract
    support.

    The client is a wrapper intended to make the Web3 library
    easier to use.

    There are three ways to use the base client class:
    1. Instantiate an object representing a blockchain and, optionally,
       a smart contract.
    2. Specialize it by making a subclass. Override the parameters
       to match the desired blockchain and, optionally, contract.
    3. If the blockchain and contract you need to use is supported,
       just use one of the 'make' methods in factory.py.

    Attributes
    ----------------------
    node_uri: str | RPC node to use
    chain_id: int = None | ID of the chain
    tx_type: int = 2 | Type of transaction: 1 for pre-EIP-1599, 2 for EIP-1599. More details here > https://docs.infura.io/infura/networks/ethereum/concepts/transaction-types
    private_key: str = None | Private key to use (optional)
    max_priority_fee_in_gwei: float = 1 | Miner's tip, relevant only for type-2 transactions (optional, default is 1)
    upper_limit_for_base_fee_in_gwei: float = inf | Raise an exception if baseFee is larger than this (optional, default is no limit)
    contract_address: Address = None | Address of smart contract (optional)
    abi: dict[str, Any] = None | ABI of smart contract; to generate from a JSON file, use static method get_contract_abi_from_file() (optional)
    middlewares: List[Middleware] = [] | Ordered list of web3.py middlewares to use (optional, default is no middlewares)


    Derived attributes
    ------------------
    w3: Web3 = None | Web3.py client
    account: LocalAccount = None | Account object for the user
    user_address: Address = None | Address of the user
    contract: Contract = None | Contract object of web3.py
    functions: ContractFunctions = None | ContractFunctions object of web3.py
    """

    def __init__(
        self,
        node_uri: str,
        chain_id: int = None,
        tx_type: int = 2,
        private_key: str = None,
        max_priority_fee_in_gwei: float = 1,
        upper_limit_for_base_fee_in_gwei: float = float("inf"),
        contract_address: Address = None,
        abi: dict[str, Any] = None,
        middlewares: List[Middleware] = [],
    ) -> None:
        # Set attributes
        self.chain_id: int = chain_id
        self.tx_type: int = tx_type
        self.max_priority_fee_in_gwei: float = max_priority_fee_in_gwei
        self.upper_limit_for_base_fee_in_gwei: float = upper_limit_for_base_fee_in_gwei
        # Initialize web3.py provider
        if node_uri:
            self.set_provider(node_uri)
        # User account
        if private_key:
            self.set_account(private_key)
        # Initialize the contract
        # TODO: we should be able to load an ABI without a specific address.
        # This might be useful to access the ABI decoding functions of web3.
        # For example, to read events from a tx only the ABI is needed, you
        # do not need the token address.
        if contract_address and abi:
            self.set_contract(contract_address, abi)
        # Add web3.py middlewares
        if middlewares:
            self.set_middlewares(middlewares)

    ####################
    # Setters
    ####################

    def set_provider(self, node_uri: str) -> None:
        self.node_uri: str = node_uri
        self.w3 = self.get_provider(node_uri)

    def set_account(self, private_key: str) -> None:
        self.private_key: str = private_key
        self.account: LocalAccount = Account.from_key(private_key)
        self.user_address: Address = self.account.address

    def set_contract(self, contract_address: Address, abi: dict[str, Any]) -> None:
        self.contract_address: Address = cast(
            Address, Web3.to_checksum_address(contract_address)
        )
        self.abi: dict[str, Any] = abi
        self.contract = self.getContract(contract_address, self.w3, abi=abi)
        self.functions = self.contract.functions

    def set_middlewares(self, middlewares: List[Middleware]) -> None:
        self.middlewares: List[Middleware] = middlewares
        for i, m in enumerate(middlewares):
            self.w3.middleware_onion.inject(m, layer=i)

    ####################
    # Build Tx
    ####################

    def build_base_tx(
        self,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: float = None,
    ) -> TxParams:
        """
        Build a basic transaction with type, nonce, chain ID and gas

        - If not given, the nonce will be computed on chain
        - If not given, the gas limit will be estimated on chain using gas_estimate()
        - For type-2 transactions, if not given, the miner's tip (maxPriorityFeePerGas)
          will be set to self.max_priority_fee_in_gwei
        - For type-2 transactions, the max gas fee is estimated according to the usual
          formula maxMaxFeePerGas = 2 * baseFee + maxPriorityFeePerGas.
        - For type-1 transactions, the gasPrice is estimated on-chain using eth_gasPrice.
        """

        # Properties that you are not likely to change
        tx: TxParams = {
            "chainId": self.chain_id,
            "from": self.user_address,
        }

        # Compute gas fee based on the transaction type
        gas_fee_in_gwei: float = None

        # Pre EIP-1599, we only have gasPrice
        if self.tx_type == 1:
            self.w3.eth.set_gas_price_strategy(rpc.rpc_gas_price_strategy)
            tx["gasPrice"] = self.w3.eth.generate_gas_price()
            gas_fee_in_gwei = float(Web3.from_wei(tx["gasPrice"], "gwei"))

        # Post EIP-1599, we have both the miner's tip and the max fee.
        elif self.tx_type == 2:
            tx["type"] = Web3.to_hex(self.tx_type)

            # The miner tip is user-provided
            max_priority_fee_in_gwei = (
                max_priority_fee_in_gwei or self.max_priority_fee_in_gwei
            )
            tx["maxPriorityFeePerGas"] = Web3.to_wei(max_priority_fee_in_gwei, "gwei")

            # The max fee is estimated from the miner tip & block base fee
            (maxFeePerGasInGwei, gas_fee_in_gwei) = self.estimate_max_fee_in_gwei(
                max_priority_fee_in_gwei
            )
            tx["maxFeePerGas"] = Web3.to_wei(maxFeePerGasInGwei, "gwei")
        else:
            raise Web3ClientException(
                f"Transaction with tx_type={self.tx_type} not supported, use either 1 or 2"
            )

        # Raise an exception if the fee is too high
        self.raise_if_gas_fee_too_high(gas_fee_in_gwei)

        # If not explicitly given, fetch the nonce on chain
        tx["nonce"] = self.get_nonce() if nonce is None else nonce

        # If needed, let web3 compute the gas-limit on chain.
        # For more details, see docs of ContractFunction.transact in
        # https://web3py.readthedocs.io/en/stable/contracts.html
        if gas_limit is not None:
            tx["gas"] = gas_limit

        return tx

    def build_tx_with_value(
        self,
        to: Address,
        value_in_eth: float,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: int = None,
    ) -> TxParams:
        """
        Build a transaction involving a transfer of value (in ETH) to an address,
        where the value is expressed in the blockchain token (e.g. ETH or AVAX).
        """
        value_in_wei = self.w3.to_wei(value_in_eth, "ether")
        return self.build_tx_with_value_in_wei(
            to, value_in_wei, nonce, gas_limit, max_priority_fee_in_gwei
        )

    def build_tx_with_value_in_wei(
        self,
        to: Address,
        value_in_wei: Wei,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: int = None,
    ) -> TxParams:
        """
        Build a transaction involving a transfer of value (in Wei) to an address,
        where the value is expressed in the blockchain token (e.g. ETH or AVAX).
        """
        tx = self.build_base_tx(
            nonce,
            gas_limit,
            max_priority_fee_in_gwei,
        )
        extra_params: TxParams = {
            "to": to,
            "value": value_in_wei,
            "gas": self.estimate_gas_for_transfer(to, value_in_wei),
        }
        return tx | extra_params

    def build_contract_tx(
        self,
        contract_function: ContractFunction,
        value_in_wei: Wei = None,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: int = None,
    ) -> TxParams:
        """
        Build a transaction that involves a contract interation.

        Requires passing the contract function as detailed in the docs:
        https://web3py.readthedocs.io/en/stable/web3.eth.account.html#sign-a-contract-transaction
        """
        base_tx = self.build_base_tx(
            nonce,
            gas_limit,
            max_priority_fee_in_gwei,
        )
        if value_in_wei:
            base_tx["value"] = value_in_wei
        return contract_function.build_transaction(base_tx)

    ####################
    # Sign & send Tx
    ####################

    def sign_tx(self, tx: TxParams) -> SignedTransaction:
        """
        Sign the give transaction; the private key must have
        been set with setCredential().
        """
        return self.w3.eth.account.sign_transaction(tx, self.private_key)

    def send_signed_tx(self, signed_tx: SignedTransaction) -> HexStr:
        """
        Send a signed transaction and return the tx hash
        """
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        return self.w3.to_hex(tx_hash)

    def sign_and_send_tx(self, tx: TxParams) -> HexStr:
        """
        Sign a transaction and send it
        """
        signed_tx = self.sign_tx(tx)
        return self.send_signed_tx(signed_tx)

    def get_tx_receipt(self, txHash: HexStr) -> TxReceipt:
        """
        Given a transaction hash, wait for the blockchain to confirm
        it and return the tx receipt.
        """
        return self.w3.eth.wait_for_transaction_receipt(txHash)

    def get_tx(self, txHash: Union[HexStr, HexBytes]) -> TxData:
        """
        Given a transaction hash, get the transaction; will raise error
        if the transaction has not been mined yet.
        """
        return self.w3.eth.get_transaction(txHash)

    def send_eth(
        self,
        to: Address,
        value_in_eth: float,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: int = None,
    ) -> HexStr:
        """
        Send ETH to the given address
        """
        tx = self.build_tx_with_value(
            to, value_in_eth, nonce, gas_limit, max_priority_fee_in_gwei
        )
        return self.sign_and_send_tx(tx)

    def send_eth_in_wei(
        self,
        to: Address,
        value_in_wei: Wei,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: int = None,
    ) -> HexStr:
        """
        Send ETH (in Wei) to the given address
        """
        tx = self.build_tx_with_value_in_wei(
            to, value_in_wei, nonce, gas_limit, max_priority_fee_in_gwei
        )
        return self.sign_and_send_tx(tx)

    ####################
    # Messages
    ####################

    def sign_message(self, msg: str) -> SignedMessage:
        """Sign the given message and return the signed message.
        NB: The method uses the 'defunct' encoding for the message, see
        https://eth-account.readthedocs.io/en/stable/eth_account.html
        for more details"""
        msg_hash = encode_defunct(text=msg)
        return self.w3.eth.account.sign_message(msg_hash, self.private_key)

    def is_message_signed_by_me(self, msg: str, signedMessage: SignedMessage) -> bool:
        """Return true if the given defunct-encoded message was signed by me"""
        msg_hash = encode_defunct(text=msg)
        signer_address = self.w3.eth.account.recover_message(
            msg_hash, signature=signedMessage.signature
        )
        return signer_address == self.user_address

    ####################
    # Scan
    ####################

    def subscribe(
        self,
        on_notification: Callable[[Any], None],
        on_subscribe: Callable[[Any], None] = None,
        once: bool = False,
        subscription_type: str = "newPendingTransactions",
    ) -> None:
        """Look for new pending transactions, blocks or events; when one is found,
        call the 'on_notification' callback.

        Notifications are processed one at a time; if you need concurrent execution,
        use async_subscribe() instead.

        Subscription types (same as eth_subscribe RPC method):

         - Use 'newHeads' to listen to new blocks. The callback receives a dict with
           the block parameters as argument.
         - Use 'newPendingTransactions' to listen to pending transactions. The callback
           receives the transaction hash as argument.
         - Use 'logs' to listen to contract event logs. The callback receives a dict
           with the smart contract address, block info and the 'data' field. [UNTESTED!]
         - For a full reference, see https://geth.ethereum.org/docs/interacting-with-geth/rpc/pubsub

        Details:

         - Provid once=True to stop the subscription after the first occurrency.
         - If you use Alchemy, you might want to use 'alchemy_newPendingTransactions'
           (https://docs.alchemy.com/reference/newpendingtransactions)
         - Subscription is good to react fast to changes on the blockchain, but you might
           miss some events. If you are ok with less but more reliable approach, use filters.

        Caveats:

         - The client must be configured with a websocket RPC endpoint (ws:// or wss://)
           or a local IPC endpoint (ending in .ipc).
         - Not all chains support the 'newPendingTransactions' subscription type.
         - Not all chains have a mempool, e.g. Arbitrum. For these chains, the function will
           just hang if asking for 'newPendingTransactions'.
         - Some chains require a validator node with staked L1 funds to be able to
           access to pending transactions while they are in the mempool (e.g. Avalanche).
           On these chains, you might still receive pending transactions, but they will
           likely be already mined.
        """
        # Raise if not a websocket uri
        rpc_url = self.node_uri
        if not rpc_url.startswith("ws") and not rpc_url.endswith(".ipc"):
            raise ValueError("Websocket RPC needed to subscribe to node notifications")

        def process_notification(
            notification: Union[str, bytes], subscription_id: str
        ) -> None:
            id, data = parse_notification(notification, subscription_type)
            if id == subscription_id:
                on_notification(data)

        async def main() -> None:
            # Connect to websocket
            async with connect(self.node_uri) as ws:
                # Subscribe to the notification type
                subscription_id = await subscribe_to_notification(
                    ws, subscription_type, on_subscribe
                )
                # Main loop
                while True:
                    # Wait for new notifications
                    notification = await asyncio.wait_for(ws.recv(), timeout=15)
                    process_notification(notification, subscription_id)
                    if once:
                        return

        asyncio.run(main())

    async def async_subscribe(
        self,
        on_notification: Callable[[Any], Awaitable[None]],
        on_subscribe: Callable[[Any], None] = None,
        once: bool = False,
        subscription_type: str = "newPendingTransactions",
    ) -> None:
        """Look for new pending transactions, blocks or events and, when one is found,
        call the 'on_notification' callback concurrently.

        For more details, see subscribe().
        """
        # Raise if not a websocket uri
        rpc_url = self.node_uri
        if not rpc_url.startswith("ws") and not rpc_url.endswith(".ipc"):
            raise ValueError("Websocket RPC needed to subscribe to node notifications")

        async def process_notification(
            notification: Union[str, bytes], subscription_id: str
        ) -> None:
            id, data = parse_notification(notification, subscription_type)
            if id == subscription_id:
                asyncio.create_task(on_notification(data))  # type: ignore

        # Connect to websocket
        async with connect(self.node_uri) as ws:
            # Subscribe to the notification type
            subscription_id = await subscribe_to_notification(
                ws, subscription_type, on_subscribe
            )
            # Main loop
            while True:
                # Wait for new notifications
                notification = await asyncio.wait_for(ws.recv(), timeout=15)
                await process_notification(notification, subscription_id)
                if once:
                    return

    ####################
    # Gas
    ####################

    def estimate_max_fee_in_gwei(
        self, max_priority_fee_in_gwei: float
    ) -> Tuple[float, float]:
        """
        For Type-2 transactions (post EIP-1559), estimate the maxFeePerGas
        parameter using the formula 2 * baseFee + maxPriorityFeePerGas.

        This is the same formula used by web3 and here
        https://ethereum.stackexchange.com/a/113373/89782

        The baseFee is fetched on chain from the latest block.

        Returns both the estimate (in gwei) and the baseFee
        (also in gwei).
        """
        latest_block = self.w3.eth.get_block("latest")
        base_fee_in_wei = latest_block["baseFeePerGas"]
        base_fee_in_gwei = float(Web3.from_wei(base_fee_in_wei, "gwei"))
        return (2 * base_fee_in_gwei + max_priority_fee_in_gwei, base_fee_in_gwei)

    def estimate_gas_price_in_gwei(self) -> float:
        """
        For Type-1 transactions (pre EIP-1559), estimate the gasPrice
        parameter by asking it directly to the node.

        Docs: https://web3py.readthedocs.io/en/stable/gas_price.html
        """
        self.w3.eth.set_gas_price_strategy(rpc.rpc_gas_price_strategy)
        return float(Web3.from_wei(self.w3.eth.generate_gas_price(), "gwei"))

    def raise_if_gas_fee_too_high(self, gas_fee_in_gwei: float) -> None:
        """
        Raise an exception if the given gas fee in Gwei is too high.

        For Type-1 transactions, pass the gasPrice; for Type-2,
        pass the baseFee.
        """
        if (
            self.upper_limit_for_base_fee_in_gwei is not None
            and gas_fee_in_gwei > self.upper_limit_for_base_fee_in_gwei
        ):
            raise TransactionTooExpensive(
                f"Gas too expensive [fee={gas_fee_in_gwei} gwei, max={self.upper_limit_for_base_fee_in_gwei} gwei]"
            )

    ####################
    # Read
    ####################

    def get_nonce(self, address: Address = None) -> Nonce:
        if not address:
            address = self.user_address
        return self.w3.eth.get_transaction_count(address)

    def get_latest_block(self) -> BlockData:
        """
        Return the latest block
        """
        return self.w3.eth.get_block("latest")

    def get_pending_block(self) -> BlockData:
        """
        Return the pending block
        """
        return self.w3.eth.get_block("pending")

    def estimate_gas_for_transfer(self, to: Address, value_in_wei: Wei) -> int:
        """
        Return the gas that would be required to send some ETH
        (expressed in Wei) to an address
        """
        return self.w3.eth.estimate_gas(
            {
                "from": self.user_address,
                "to": to,
                "value": value_in_wei,
            }
        )

    def get_balance_in_wei(self, address: Address = None) -> Wei:
        return self.w3.eth.get_balance(Web3.to_checksum_address(address))

    def get_balance_in_eth(self, address: Address = None) -> float:
        return float(Web3.from_wei(self.get_balance_in_wei(address), "ether"))

    ####################
    # Contract
    ####################

    def call(
        self,
        function: ContractFunction,
        tx: TxParams = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride = None,
    ) -> Any:
        """
        Execute a contract function call using the eth_call interface.
        This won't write to the blockchain.

        Example: get the balance of an address for an ERC20 token:
            client.call(
                client.functions.balanceOf(address)
            )
        """
        return function.call(tx, block_identifier, state_override)

    def transact(
        self,
        function: ContractFunction,
        value_in_wei: Wei = None,
        nonce: Nonce = None,
        gas_limit: int = None,
        max_priority_fee_in_gwei: int = None,
    ) -> HexStr:
        """
        Execute a contract function.
        This will write to the blockchain.

        Example: transfer some tokens to the given address:
            client.transact(
                client.functions.transfer(address, amount)
            )
        """
        tx: TxParams = self.build_contract_tx(
            function, value_in_wei, nonce, gas_limit, max_priority_fee_in_gwei
        )
        return self.sign_and_send_tx(tx)

    ####################
    # Utils
    ####################

    @staticmethod
    def getContract(
        address: Address,
        provider: Web3,
        abi_file: str = None,
        abi: dict[str, Any] = None,
    ) -> Contract:
        """
        Load the smart contract, required before running
        build_contract_tx().
        """
        checksum = Web3.to_checksum_address(address)
        if abi_file:
            abi = BaseClient.get_contract_abi_from_file(abi_file)
        return provider.eth.contract(address=checksum, abi=abi)

    @staticmethod
    def get_contract_abi_from_file(file_name: str) -> Any:
        with open(file_name, encoding="utf-8") as file:
            return json.load(file)

    @staticmethod
    def get_provider(node_uri: str) -> Web3:
        """
        Initialize provider (HTTPS & WS supported).

        TODO: Support autodetection with empty node_uri
        docs here https://web3py.readthedocs.io/en/stable/providers.html#how-automated-detection-works
        """
        if node_uri[0:4] == "http":
            return Web3(Web3.HTTPProvider(node_uri))
        elif node_uri[0:2] == "ws":
            return Web3(Web3.WebsocketProvider(node_uri))
        else:
            return Web3()

    @staticmethod
    def get_gas_spent_in_eth(txReceipt: TxReceipt) -> float:
        """
        Given the transaction receipt, return the ETH that
        was spent in gas to process the transaction
        """
        return float(
            Web3.from_wei(
                txReceipt["effectiveGasPrice"] * txReceipt["gasUsed"], "ether"
            )
        )



from typing import Dict

from web3client.base_client import BaseClient


def test_get_nonce(address: str, networks_clients: Dict[str, BaseClient]) -> None:
    for network, client in networks_clients.items():
        nonce = client.get_nonce(address)
        assert type(nonce) is int
        assert nonce >= 0


def test_get_latest_block(networks_clients: Dict[str, BaseClient]) -> None:
    for network, client in networks_clients.items():
        block = client.get_latest_block()
        assert type(block.get("number")) is int
        assert block.get("number") >= 0
        assert type(block.get("size")) is int
        assert block.get("size") >= 0
        assert type(block.get("difficulty")) is int
        assert block.get("difficulty") >= 0
        assert type(block.get("transactions")) is list


def test_get_eth_balance(address: str, networks_clients: Dict[str, BaseClient]) -> None:
    for network, client in networks_clients.items():
        balance = client.get_balance_in_wei(address)
        assert type(balance) is int
        assert balance >= 0


def test_get_sign_message(
    private_key: str, networks_clients: Dict[str, BaseClient]
) -> None:
    msg = "Hello world!"
    for network, client in networks_clients.items():
        client.set_account(private_key=private_key)
        signed_message = client.sign_message(msg)
        assert client.is_message_signed_by_me(msg, signed_message)



"""
PyTest Fixtures.
"""

import json
from pathlib import Path

import pytest
from web3.types import ABI

import ape


@pytest.fixture()
def simple_abi() -> ABI:
    """A simple ABI for a contract with a single function"""
    return [
        {
            "constant": False,
            "inputs": [{"name": "a", "type": "uint256"}],
            "name": "foo",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        }
    ]


@pytest.fixture()
def erc20_abi_string() -> str:
    """The ABI for the ERC20 token standard, as a string"""
    return '[{"constant":true,"inputs":[],"name":"name","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_spender","type":"address"},{"name":"_value","type":"uint256"}],"name":"approve","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"totalSupply","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transferFrom","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"decimals","outputs":[{"name":"","type":"uint8"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"}],"name":"balanceOf","outputs":[{"name":"balance","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"symbol","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transfer","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"},{"name":"_spender","type":"address"}],"name":"allowance","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"payable":true,"stateMutability":"payable","type":"fallback"},{"anonymous":false,"inputs":[{"indexed":true,"name":"owner","type":"address"},{"indexed":true,"name":"spender","type":"address"},{"indexed":false,"name":"value","type":"uint256"}],"name":"Approval","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"from","type":"address"},{"indexed":true,"name":"to","type":"address"},{"indexed":false,"name":"value","type":"uint256"}],"name":"Transfer","type":"event"}]'


@pytest.fixture()
def erc20_abi_file(tmp_path: Path, erc20_abi_string: str) -> str:
    """The path of a JSON file containing the ABI for the ERC20 token
    standard"""
    f = tmp_path / "erc20.json"
    f.write_text(erc20_abi_string)
    return str(f)


@pytest.fixture()
def erc20_abi(erc20_abi_string: str) -> ABI:
    """The ABI for the ERC20 token standard, as a JSON object"""
    return json.loads(erc20_abi_string)


@pytest.fixture()
def alice() -> ape.api.accounts.AccountAPI:
    """An ape account preloaded in the local chain"""
    return ape.accounts[0]


@pytest.fixture()
def bob() -> ape.api.accounts.AccountAPI:
    """An ape account preloaded in the local chain"""
    return ape.accounts[1]



from functools import reduce
from typing import Any, List
from collections import Counter
from random import uniform


def firstOrNone(list: List[Any]) -> Any:
    """
    Return the first element of a list or None if it is not set
    """
    return nthOrNone(list, 0)


def secondOrNone(list: List[Any]) -> Any:
    """
    Return the second element of a list or None if it is not set
    """
    return nthOrNone(list, 1)


def thirdOrNone(list: List[Any]) -> Any:
    """
    Return the third element of a list or None if it is not set
    """
    return nthOrNone(list, 2)


def fourthOrNone(list: List[Any]) -> Any:
    """
    Return the fourth element of a list or None if it is not set
    """
    return nthOrNone(list, 3)


def nthOrNone(list: List[Any], n: int) -> Any:
    """
    Return the n-th plus 1 element of a list or None if it is not set
    """
    try:
        return list[n]
    except:
        return None


def nthOrLastOrNone(list: List[Any], n: int) -> Any:
    """
    Return the n-th element of a list; if it is not set, return
    the last element of the list; if it is not set, return none.
    """
    if not list:
        return None
    return list[n] if len(list) > n else list[-1]


def findInListOfDicts(l: List[dict[str, Any]], key: str, value: Any) -> Any:
    """
    Return the first dictionary in the list that has the
    given key value
    """
    return firstOrNone([item for item in l if item[key] == value])


def indexInList(l: List[Any], value: Any, doPop: bool = False) -> int:
    """
    Wrapper to the list.index(value) method which returns None
    if the value is not found, instead of raising an exception.
    """
    try:
        i = l.index(value)
        if doPop:
            l.pop(i)
        return i
    except ValueError:
        return None


def duplicatesInList(l: List[Any]) -> List[Any]:
    """
    Return duplicate elements in the given list

    Source: https://stackoverflow.com/a/9835819/2972183
    """
    return [item for item, count in Counter(l).items() if count > 1]


def flattenList(l: List[Any]) -> List[Any]:
    """
    Flatten a list

    Source: https://stackoverflow.com/a/46080186/2972183
    """
    return reduce(lambda x, y: x + y, l)


def randomize(x: float, d: float) -> float:
    """
    Return a number uniformly distributed between
    x*(1-d) and x*(1+d)
    """
    if d == 0:
        return x
    return x * (1 + uniform(-1, 1) * d)



from random import choice
from typing import Any, List, cast

from web3.middleware import geth_poa_middleware

from web3client.exceptions import NetworkNotFound
from web3client.helpers.general import findInListOfDicts
from web3factory.types import NetworkConfig

"""
List of supported networks (aka blockchains), each with its own
parameters.
"""
supported_networks: List[NetworkConfig] = [
    # Ethereum
    {
        "name": "eth",
        "tx_type": 2,
        "chain_id": 1,
        "rpcs": [
            "https://cloudflare-eth.com",
        ],
        "coin": "ETH",
    },
    # Avalanche C Chain
    {
        "name": "bnb",
        "tx_type": 1,
        "chain_id": 56,
        "middlewares": [geth_poa_middleware],
        "rpcs": [
            "https://bsc-dataseed.binance.org/",
        ],
        "coin": "BNB",
    },
    # Avalanche C Chain
    {
        "name": "avax",
        "tx_type": 2,
        "chain_id": 43114,
        "middlewares": [geth_poa_middleware],
        "rpcs": [
            "https://api.avax.network/ext/bc/C/rpc",
        ],
        "coin": "AVAX",
    },
    # Arbitrum One
    {
        "name": "arb",
        "tx_type": 2,
        "chain_id": 42161,
        "rpcs": ["https://arb1.arbitrum.io/rpc"],
        "coin": "ETH",
    },
    # zkSync Era
    {
        "name": "era",
        "tx_type": 2,
        "chain_id": 324,
        "rpcs": ["https://mainnet.era.zksync.io"],
        "coin": "ETH",
    },
]


def get_network_config(name: str) -> NetworkConfig:
    """
    Return the configuration for the network with the given
    name; raises an exception if not found
    """
    network: NetworkConfig = findInListOfDicts(
        cast(Any, supported_networks), "name", name
    )
    if network is None:
        raise NetworkNotFound(f"Network '{name}' not supported")
    return network


def pick_random_rpc(network_name: str) -> str:
    """
    Given a network return one of its RPCs, randomly,
    or None, if it has no RPC
    """
    network = get_network_config(network_name)
    rpcs = network.get("rpcs")
    return choice(rpcs) if rpcs else None


def pick_first_rpc(network_name: str) -> str:
    """
    Given a network return its first RPC or None,
    if it has no RPC
    """
    network = get_network_config(network_name)
    rpcs = network.get("rpcs")
    return rpcs[0] if rpcs else None


def is_network_supported(name: str) -> bool:
    """
    Return true if the given network is supported by
    the client factory
    """
    try:
        get_network_config(name)
        return True
    except NetworkNotFound:
        return False



import json
from typing import Any, Callable, Tuple, Union

from websockets.client import WebSocketClientProtocol

from web3client.exceptions import Web3ClientException


async def subscribe_to_notification(
    ws: WebSocketClientProtocol,
    type: str,
    on_subscribe: Callable[[Any], None] = None,
) -> str:
    """Given a websocket connection to an RPC, subscribe to the given
    notification type, and return the subscription id.

    Optinally, call the given callback with the subscription response."""
    # Subscribe to the notification type
    await ws.send(
        '{"jsonrpc": "2.0", "id": 1, "method": "eth_subscribe", "params": ["'
        + type
        + '"]}'
    )
    subscription_response = await ws.recv()
    try:
        subscription_id = json.loads(subscription_response)["result"]
    except Exception as e:
        raise Web3ClientException(f"Failed to subscribe to {type}: {e}")
    # Call on_subscribe callback
    if on_subscribe:
        on_subscribe(json.loads(subscription_response))
    return subscription_id


def parse_notification(notification: Union[str, bytes], type: str) -> Tuple[str, Any]:
    """Given a notification, return the subscription ID and
    notification data"""
    try:
        as_dict = json.loads(notification)
    except json.JSONDecodeError as e:
        raise Web3ClientException(
            f"Notification from websocket is malformed: {notification!r}"
        )

    try:
        subscription_id = as_dict["params"]["subscription"]
    except KeyError:
        raise Web3ClientException(
            f"Cannot extract 'subscription' field from websocket notification: {notification!r}"
        )

    try:
        data = as_dict["params"]["result"]
    except KeyError:
        raise Web3ClientException(
            f"Cannot extract 'result' field from websocket notification: {notification!r}"
        )

    return subscription_id, data



from typing import Any, Type, cast

from eth_typing import Address

from web3client.base_client import BaseClient
from web3client.erc20_client import Erc20Client
from web3factory.erc20_tokens import get_token_config
from web3factory.networks import get_network_config, pick_first_rpc
from web3factory.types import NetworkName, TokenName


def make_client(
    networkName: NetworkName,
    node_uri: str = None,
    base: Type[BaseClient] = BaseClient,
    **clientArgs: Any,
) -> BaseClient:
    """
    Return a brand new client configured for the given blockchain
    """
    networkConfig = get_network_config(networkName)
    if node_uri is None:
        node_uri = pick_first_rpc(networkName)
    client = base(node_uri=node_uri, **clientArgs)
    client.chain_id = networkConfig["chain_id"]
    client.tx_type = networkConfig["tx_type"]
    client.set_middlewares(networkConfig.get("middlewares", []))

    return client


def make_erc20_client(
    tokenName: TokenName,
    networkName: NetworkName,
    node_uri: str = None,
    **clientArgs: Any,
) -> Erc20Client:
    """
    Return a brand new client configured for the given blockchain
    and preloaded with the ERC20 token ABI
    """
    token_address = get_token_config(tokenName, networkName)["address"]
    client = make_client(
        networkName, node_uri, Erc20Client, contract_address=token_address, **clientArgs
    )
    return cast(Erc20Client, client)



from typing import List, TypedDict

from typing_extensions import NotRequired
from web3.types import Middleware

NetworkName = str
TokenName = str


class NetworkConfig(TypedDict):
    """
    Dictionary representing the configuration of a network, e.g. Ethereum,
    Avalanche, etc.
    """

    name: NetworkName
    tx_type: int
    chain_id: int
    middlewares: NotRequired[List[Middleware]]
    rpcs: NotRequired[List[str]]
    coin: str


class Erc20TokenConfig(TypedDict):
    """
    Dictionary representing an ERC20 token
    """

    name: TokenName
    network: NetworkName
    address: str
    decimals: int



from typing import List

from web3client.exceptions import (
    Erc20TokenNotFound,
    Erc20TokenNotUnique,
    NetworkNotFound,
)
from web3factory.networks import is_network_supported
from web3factory.types import Erc20TokenConfig

"""
List of supported ERC20 tokens acrosso networks
"""
supported_tokens: List[Erc20TokenConfig] = [
    # Ethereum
    {
        "name": "USDC",
        "network": "eth",
        "address": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
        "decimals": 6,
    },
    # Binance
    {
        "name": "BUSD",
        "network": "bnb",
        "address": "0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56",
        "decimals": 18,
    },
    {
        "name": "BETH",
        "network": "bnb",
        "address": "0x2170Ed0880ac9A755fd29B2688956BD959F933F8",
        "decimals": 18,
    },
    # Avalanche
    {
        "name": "USDC",
        "network": "avax",
        "address": "0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E",
        "decimals": 6,
    },
]


def get_token_config(name: str, network: str) -> Erc20TokenConfig:
    """
    Return the configuration for the given token on the given
    network; raises an exception if not found or more than one
    is found
    """
    # Raise if network does not exist
    if not is_network_supported(network):
        raise NetworkNotFound(f"Network '{network}' not supported")
    # Get all tokens with given name and network
    tokens: List[Erc20TokenConfig] = [
        t for t in supported_tokens if t["name"] == name and t["network"] == network
    ]
    # Must have exactly one token
    if len(tokens) == 0:
        raise Erc20TokenNotFound(f"ERC20 token '{name}' on '{network}' not supported")
    if len(tokens) > 1:
        raise Erc20TokenNotUnique(
            f"Found more than one ERC20 token with '{name}' on '{network}' (found '{len(tokens)}')"
        )
    # Return the one token
    return tokens[0]


def is_token_supported(name: str, network: str) -> bool:
    """
    Return true if the given token and network pair is
    supported by the client factory
    """
    try:
        get_token_config(name, network)
        return True
    except Erc20TokenNotFound:
        return False



class Web3ClientException(BaseException):
    pass


class MissingParameter(Web3ClientException):
    pass


class TransactionTooExpensive(Web3ClientException):
    pass


class NetworkNotFound(Web3ClientException):
    pass


class Erc20TokenNotFound(Web3ClientException):
    pass


class Erc20TokenNotUnique(Web3ClientException):
    pass



import os

from eth_typing import Address, HexStr
from web3 import Web3
from web3.types import Nonce, TxParams, Wei

from web3client.base_client import BaseClient


class Erc20Client(BaseClient):
    """
    Client that comes with the ERC20 ABI preloaded.

    Call the contract functions directly: transfer, balanceOf, name,
    etc.

    AMOUNTS
    =======

    Whenever we will refer to an "amount" of the token, we really mean an
    "amount in token units". A token unit is the smallest subdivision of
    the token. For example:
    - If the token has 6 digits (like most stablecoins) an amount of 1
      corresponds to one millionth of the token.
    - For tokens with 18 digits (like most non-stablecoins) an amount
      of 1 is equal to 1/10^18 of the token (a single wei).
    """

    abiDir = os.path.dirname(os.path.realpath(__file__)) + "/abi"
    abi = BaseClient.get_contract_abi_from_file(abiDir + "/erc20.json")

    def __init__(
        self,
        node_uri: str,
        chain_id: int = None,
        tx_type: int = 2,
        private_key: str = None,
        max_priority_fee_in_gwei: float = 1,
        upper_limit_for_base_fee_in_gwei: float = float("inf"),
        contract_address: Address = None,
    ) -> None:
        super().__init__(
            node_uri,
            chain_id,
            tx_type,
            private_key,
            max_priority_fee_in_gwei,
            upper_limit_for_base_fee_in_gwei,
            contract_address,
            self.abi,
        )

    ####################
    # Read
    ####################

    def balanceOf(self, address: Address) -> int:
        """
        Return the amount held by the given address
        """
        return self.contract.functions.balanceOf(
            Web3.to_checksum_address(address)
        ).call()

    def name(self) -> str:
        """
        Return the name/label of the token
        """
        return self.contract.functions.name().call()

    def symbol(self) -> str:
        """
        Return the symbol/ticker of the token
        """
        return self.contract.functions.symbol().call()

    def totalSupply(self) -> int:
        """
        Return the total supply of the token
        """
        return self.contract.functions.totalSupply().call()

    def decimals(self) -> int:
        """
        Return the number of digits of the token
        """
        return self.contract.functions.decimals().call()

    ####################
    # Write
    ####################

    def transfer(
        self, to: Address, amount: int, nonce: Nonce = None, value_in_wei: Wei = Wei(0)
    ) -> HexStr:
        """
        Transfer some amount of the token to an address; does not
        require approval.
        """

        tx: TxParams = self.build_contract_tx(
            self.contract.functions.transfer(Web3.to_checksum_address(to), amount)
        )

        if nonce:
            tx["nonce"] = nonce

        if value_in_wei:
            tx["value"] = value_in_wei

        return self.sign_and_send_tx(tx)

    ####################
    # Static
    ####################

    @staticmethod
    def from_wei(amount: int, decimals: int) -> float:
        """
        Given an amount in Wei, return the equivalent amount in
        ETH units
        """
        return amount / 10**decimals



from typing import Dict

import pytest

from web3client.base_client import BaseClient
from web3factory.factory import make_client
from web3factory.networks import supported_networks


@pytest.fixture()
def rpcs() -> Dict[str, str]:
    """
    Let's use difrerent RPCs for tests, in case the regular ones
    are throttled
    """
    return {
        "eth": "https://mainnet.infura.io/v3/db0e363aad2f43ee8a2f259733721512",
    }


@pytest.fixture()
def address() -> str:
    return "0x3A8c8833Abe2e8454F59574A2A18b9bA8A28Ea4F"


@pytest.fixture()
def private_key() -> str:
    return "53caa63985c6089c84be07e3f42d5d7ebd47a8a097835ede937d4c5e1f1021dd"


@pytest.fixture()
def networks_clients(rpcs: Dict[str, str]) -> Dict[str, BaseClient]:
    """
    Ready-to-use clients, indexed by network name, no signer
    """
    clients = {}
    for network in supported_networks:
        name = network["name"]
        node_uri = rpcs.get(name)
        clients[name] = make_client(name, node_uri)
    return clients


@pytest.fixture()
def eth_client() -> BaseClient:
    return make_client("eth")


@pytest.fixture()
def bnb_client() -> BaseClient:
    return make_client("bnb")


@pytest.fixture()
def bnb_ws_client() -> BaseClient:
    return make_client("bnb", node_uri="wss://dex.binance.org/api/ws")



import pprint
from typing import Any, Union

from eth_typing.encoding import HexStr
from web3.datastructures import AttributeDict
from web3.types import TxReceipt

from web3client.base_client import BaseClient


def printTxInfo(client: BaseClient, txHash: HexStr) -> None:
    """
    Get a transaction receipt and print it, together with
    the tx cost
    """
    print(">>> TX SENT!")
    print("Hash = " + txHash)
    print("Waiting for transaction to finalize...")
    tx_receipt = client.get_transaction_receipt(txHash)
    print(">>> TX IS ON THE BLOCKCHAIN :-)")
    pprint.pprint(tx_receipt)
    print(">>> ETH SPENT")
    print(BaseClient.get_gas_spent_in_eth(tx_receipt))


def pprintAttributeDict(
    attributeDict: Union[TxReceipt, AttributeDict[str, Any]]
) -> None:
    """
    Web3 often returns AttributeDict instead of simple Dictionaries;
    this function pretty prints an AttributeDict
    """
    print(formatAttributeDict(attributeDict))


def formatAttributeDict(
    attributeDict: Union[TxReceipt, AttributeDict[str, Any]],
    indent: int = 4,
    nestLevel: int = 0,
) -> str:
    """
    Web3 often returns AttributeDict instead of simple Dictionaries;
    this function return a pretty string with the AttributeDict content
    """
    prefix = nestLevel * indent * " "
    output = prefix + "{\n"
    for key, value in attributeDict.items():
        if isinstance(value, AttributeDict):
            output += prefix + formatAttributeDict(value, indent, nestLevel + 1)
            output += "\n"
        else:
            output += prefix + " " * indent
            output += f"{key} -> {pprint.pformat(value, indent=indent)}"
            output += "\n"
    output += prefix + "}"

    return output
